
;; Lista form zmikszonych spgosek tematycznych 1. declinacji.
;; UWAGA: Sortowana wg. dugoci cdr-a
(defvar lenition-1st-decl
  '(("r" . "rz")
    ("zd" . "dzi")
    ("d" . "dzi")
    ("z" . "zi")
    ("s" . "si")
    ("" . "l")
    ("st" . "ci")
    ("t" . "ci")))

(defun get-root-lenited (lenition-list word ignore-len)
  "Zwraca zmikszon form tematu rzeczownika WORD wg. listy LENITION-LIST.
Ignoruje kocowe IGNORE-LEN liter."
  (dolist (x lenition-list)
    (when (match-ending word (cdr x) ignore-len)
      (return (string-concat (substring word 0
					(- (length word) (length (cdr x))
					   ignore-len))
			     (car x))))))

(defun replace-lenition (word)
  (cond
   ((match-ending word "i")
    (let* ((c (aref word (- (length word) 2)))
	   (r (cond
	       ((char= c #\n)  "")
	       ((char= c #\s)  "")
	       ((char= c #\c)  "")
	       ((char= c #\z)  ""))))
      (if r
	  (string-concat
	   (substring word 0 (- (length word) 2))
	   r))))
   ((match-ending word "e")
    (cond
     ((match-ending word "ce")
      (string-concat
       (substring word 0 (- (length word) 2))
       "k"))
      ((match-ending word "dze")
       (string-concat
	(substring word 0 (- (length word) 3))
	"g"))
      (t
       (substring word 0 (1- (length word))))))
   (t
    word)))
	     

;; Funkcje analizujce dla poszczeglnych deklinacji/kocwek
;; Kada funkcja zwraca list list:
;;
;;   (temat forma-sownikowa deklinacja przypadek liczba)
;; 

(defun parse-on (word root ending-length)
  "Obsuga rzeczownikw 2iej dekl. typu 'imi', 'ciel' w D. l.m."
  (list (string-concat root "en") ; Zgodnie ze sowiask etymologi.
	(string-concat root "") 2 'D 'LM))

;; Jeli ROOT jest 'rdzeniem' rzeczownika 1-ej deklinacji z 'ie', zwraca form mianownika
(defun find-1dcl-root (root)
  (let ((n (assqf root '(("ojc" . "ojciec")
			 ("pan" . "pan")
			 ("chopc" . "chopiec")
			 ("koc" . "koniec")
			 ("brat" . "brat")
			 ("ps" . "pies")
			 ("kot" . "kot")
			 ("lw" . "lew")
			 ("or" . "orze")
			 ("wiat" . "wiat")
			 ("bog" . "bg")
			 ("diab" . "diabe"))
		  #'string-equal)))
    (if n
	(cdr n))))

(defvar substantive-drivers)
(setq substantive-drivers
;;    Kocwka              Lista funkcji analizucych
  (list
   ;; Obsuga rzeczownikw 2iej dekl. typu 'museum' we wszystkich
   ;; przypadkach l.p.
   (list "um"
	 #'(lambda (word root ending-length)
	     (list root word 2 '(M D C B W N Mc) 'LP)))
   ;;
   (list "on"                  #'parse-on)
   (list "t"                  #'parse-on)
   ;; 
   (list "j"
	 ;; Rzeczowniki mikkotematowe poyczone 3. deklinacji w dopeniaczu
	 ;; liczby mnogiej
	 #'(lambda (word root ending-length)
	     (cond
	      ((match-ending root "i") ;; linij -> linia
	       (list root
		     (string-concat root "a")
		     3
		     'D
		     'LM))
	      ((match-ending root "y") ;; lekcyj -> lekcja
	       (list (remove-ending root 1)
		     (string-concat (remove-ending root 1) "ja")
		     3
		     'D
		     'LM))
	      ;; FIXME: t?
	      )))
   ;;
   (list "owi"
	 ;; FIXME: narodowi -> nard
	 #'(lambda (word root ending-length)
	     (let ((x (replace-lenition root)))
	       (and x
		    (list x
			  x
			  1
			  'C
			  'LP)))))
   ;;
   (list "em"
	 #'(lambda (word root ending-length)
	     (let ((r (find-1dcl-root root)))
	       (when r
		 (list
		  root
		  r
		  1
		  'N
		  'LP))))
	 ;; Kocwk -em maj rzeczowniki 1. i 2. deklinacji w narzdniku
	 ;; liczby pojedynczej.
	 ;; Po spgoskach -k i -g zachodzi zmikszenie
	 ;; (robotnik=>robotnikiem)
	 ;; Jeli przed kocwk -em stoi przyrostek -eni- lub -ci-, moe to
	 ;; by rzeczownik 2. deklinacji z tematem na noswk (imi=>imieniem,
	 ;; zwierz=>zwierzciem), ale nie koniecznie (yczenie => yczeniem,
	 ;; pojcie => pojciem).
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "eni") 
		       (match-ending root "ci"))
	       (list
		(remove-ending word 1)
		(remove-ending word 1)
		2
		'N
		'LP)))
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "eni") 
		       (match-ending root "ci"))
	       (let ((x (remove-ending root 3)))
		 (list
		  (string-concat x "en")
		  (string-concat x "")
		  2
		  'N
		  'LP))))
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "ni")
		       (match-ending root "l"))
	       (list
		(string-concat root "e")
		(string-concat root "e")
		2
		'N
		'LP)))
	 
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "ki")
		       (match-ending root "gi"))
	       (let ((x (remove-ending root 1)))
		 (list
		  x
		  x
		  1
		  'N
		  'LP))))
	 ;; FIXME: pc, rc (chopiec, dworzec, itp.)
	 ;; gob
	 #'(lambda (word root ending-length)
	     (let ((x (replace-lenition root)))
	       (when x
		 (list
		  x
		  x
		  1
		  'N
		  'LP)))))
   (list "a"
	 ;; Kocwk -a mog miec rzeczowniki:
	 ;;   1. dekl. w dopeniaczu lp. -- chop-a, koni-a
	 #'(lambda (word root ending-length)
	     (let ((x (replace-lenition root)))
	       (when x
		 (let ((r (or (find-1dcl-root x) x)))
		   (list
		    x
		    r
		    1
		    'D
		    'LP)))))

	 ;;   1. dekl. w bierniku lp. (rzeczowniki ywotni) 
	 #'(lambda (word root ending-length)
	     (let ((x (replace-lenition (remove-ending word 1))))
	       (when x
		 (let ((r (or (find-1dcl-root x) x)))
		   (list
		    x
		    r
		    1
		    'B
		    'LP)))))

	 ;;   2. dekl. w dopeniaczu lp. -- okn-a, zdani-a, pol-a
	 #'(lambda (word root ending-length)
	     (list
	      root
	      (cond
	       ((match-ending root "i")
		(string-concat root "e"))
	       ((match-ending root "l")
		(string-concat root "e"))
	       (t
		(string-concat root "o")))
	      2
	      'D
	      'LP))

	 ;;   2. dekl. z tematem na noswk w dopeniaczu lp. -- imienia 
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "eni") 
		       (match-ending root "ci"))
	       (let ((x (remove-ending root 3)))
		 (list
		  (string-concat x "en")
		  (string-concat x "")
		  2
		  'D
		  'LP))))
	 #'(lambda (word root ending-length)
	     (when (match-ending root "ani")
	       (let ((x (remove-ending root 3)))
		 (list
		  (string-concat x "a")
		  (string-concat x "anie")
		  2
		  'D
		  'LP))))
	 
	 ;;   2. dekl. w mianowniku, bierniku bd woaczu lm. --
	 ;;    okn-a, imiona, muzea
	 #'(lambda (word root ending-length)
	     (list 
	      root
	      (string-concat root "o")
	      2
	      '(M B W)
	      'LM))

	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "on") 
		       (match-ending root "t"))
	       (let ((x (remove-ending root 2)))
		 (list
		  (string-concat x "en")
		  (string-concat x "")
		  2
		  '(M B W)
		  'LM))))

	 #'(lambda (word root ending-length)
	     (list
	      root
	      (string-concat root "um")
	      2
	      '(M B W)
	      'LM))
  
	 ;;   3. dekl. w mianowniku lp. -- on-a, ksik-a
	 #'(lambda (word root ending-length)
	     (list root word 3 'M 'LP)))
   ;;
   (list "u"
	 ;; Kocwk -u maj rzeczowniki:
	 ;;  1. deklinacji w dopeniaczu liczby pojedynczej wzoru twardego
	 #'(lambda (word root ending-length)
	     (when (hard-ending root)
	       (list root root 1 'D 'LP)))
	 
	 ;;  1. deklinacji w celowniku lp. wt. -- tylko niektre sowa
	 #'(lambda (word root ending-length)
	     (let ((r (find-1dcl-root root)))
	       (when r
		 (list root r 1 'C 'LP))))
	     
	 ;;  1. deklinacji w celowniku, woaczu i miejscowniku liczby
	 ;;  pojedynczej wzoru mikkiego
	 #'(lambda (word root ending-length)
	     (when (soft-ending root)
	       (list root
		     (replace-lenition root) 1 '(C W Mc) 'LP)))
	 
	 ;;  2. deklinacji w celowniku lp. (ewentualnie odpowiadajce
	 ;;  mianownikowi na -).
	 ;;  2. deklinacji na -'e, -, -o o temacie na spgosk
	 ;;  tylnojzykow w miejscowniku: o jabku, o polu, o imieniu,
	 ;;  o cielciu
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "eni")
		       (match-ending root "ci"))
	       (list
		(string-concat (remove-ending root 3) "en")
		(string-concat (remove-ending root 3) "")
		2
		'(C Mc)
		'LP)))
	 #'(lambda (word root ending-length)
	     (list
	      root
	      (string-concat root "o")
	      2
	      '(C Mc)
	      'LP))
	 #'(lambda (word root ending-length)
	     (list
	      root
	      (string-concat root "e")
	      2
	      '(C Mc)
	      'LP))
	 ;;  3. deklinacji liczby pojedynczej wzoru mikkiego - tylko
	 ;;  rzeczowniki zdrobniae - w wolaczu: babciu.
	  #'(lambda (word root ending-length)
	      (when (or (match-ending word "ciu")
			(match-ending word "siu"))
		(list
		 (replace-lenition root)
		 (string-concat root "a")
		 3
		 'W
		 'LP))))

   (list "e"
	 ;; Rzeczowniki 1 deklinacji w:
	 ;;   woaczu i miejscowniku liczby pojedynczej wzoru twardego --
	 ;;   ze zmiekszeniem poprzedniej spgoski
	 #'(lambda (word root ending-length)
	     (let ((r (or (get-root-lenited lenition-1st-decl word 1)
			  (remove-ending root 1))))
	       (when (hard-ending r)
		 (list r r 1 '(W Mc) 'LP))))
   
	 ;; Rzeczowniki 1 deklinacji w:
	 ;;   mianowniku, bierniku i woaczu liczby mnogiej wzoru mikkiego
	 #'(lambda (word root ending-length)
	     (when (match-ending root "i")
	       (let ((r (replace-lenition root)))
		 (when (and r (soft-ending r))
		   (list
		    r
		    r
		    1
		    '(M B W)
		    'LM))))) 
	 
	 ;; Rzeczowniki 2. deklinacji na -o o temacie na spgosk tward,
	 ;; oprcz tylnojzykowych ([7]) w miejscowniku lp.
	 #'(lambda (word root ending-length)
	     (when (and (match-ending root "i")
			(not (or (match-ending root "ki")
				 (match-ending root "gi"))))
	       (let ((r (remove-ending root 1)))
		 (list
		  r
		  (string-concat r "o")
		  2
		  'Mc
		  'LP))))
	 
	 ;; Rzeczowniki 3. deklinacji w:
	 ;;   miejscowniku, bierniku i woaczu liczby mnogiej wzoru mikkiego
	 ;;   ([17])
	 ;;; 1: sowa typu "ziemia", "linia"
	 #'(lambda (word root ending-length)
	     (when (match-ending root "i")
	       (list
		(replace-lenition (remove-ending root 1))
		(string-concat root "a")
		3
		'(Mc B W)
		'LM)))
	 ;;; sowa typu "gospodyni", "pani"
	 #'(lambda (word root ending-length)
	     (when (match-ending root "i")
	       (list
		(replace-lenition (remove-ending root 1))
		root
		3
		'(M B W)
		'LM))) 
	 
	 ;; Rzeczowniki 3. deklinacji w celowniku i miejscowniku
	 ;; liczby pojedynczej wzoru twardego. [11]
	 #'(lambda (word root ending-length)
	     (let ((r (if (and (match-ending root "i")
			       (hard-ending root))
			  (replace-lenition (remove-ending root 1))
			(replace-lenition word))))
	       (list
		r
		(string-concat r "a")
		3
		'(C Mc)
		'LP)))
	 
         ;; Rzeczowniki 4 deklinacji w:
	 ;;   miejscowniku, bierniku i woaczu liczby mnogiej [20]
	 #'(lambda (word root ending-length)
	     (let ((r (replace-lenition root)))
	       (list
		r
		r
		4
		'(Mc B W)
		'LM))))

   (list "ami"
	 ;; Rzeczowniki wszystkich deklinacji w narzdniku liczby mnogiej
	 ;;; 1. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      root
	      (replace-lenition root)
	      1
	      'N
	      'LM))
	 #'(lambda (word root ending-length)
	     (when (match-ending root "c")
	       (list
		root
		(string-concat (remove-ending root 1) "iec")
		1
		'N
		'LM)))
	 ;;; 2. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      2
	      'C
	      'LM))	     
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "on")
		       (match-ending root "t"))
	       (let ((r (string-concat
			 (remove-ending root 2)
			 "")))
		 (list
		  r
		  r
		  2
		  'N
		  'LM))))
	 ;;; 3. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      (string-concat root "a")
	      (string-concat root "a")
	      3
	      'N
	      'LM))
	 #'(lambda (word root ending-length)
	     (list
	      root 
	      root 
	      3
	      'N
	      'LM))
	 ;;; 4. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      (replace-lenition root)
	      (replace-lenition root)
	      4
	      'N
	      'LM)))
	 
   (list "mi"
	 ;; Niektre rzeczowniki 1. dekl. w narzdniku liczby mnogiej
	 ;; wzoru twardego po spgosce mikkiej
	 #'(lambda (word root ending-length)
	     (when (and (consonant-ending root)
			(soft-ending root))
	       (list
		root
		root
		1
		'N
		'LM))))
   
   (list "i"
	 ;; Rzeczowniki 1. deklinacji
	 ;;  w mianowniku i woaczu liczby mnogiej wzoru twardego. [2]
	 #'(lambda (word root ending-length)
	     (when (or (hard-ending root)
		       (match-ending root "k")
		       (match-ending root "g"))
	       (list
		root
		root
		1
		'(M W)
		'LM)))
	 ;;  w dopeniaczu liczby mnogiej wzoru mikkiego [5] po spgoskach
	 ;;  mikkich
	 #'(lambda (word root ending-length)
	     (let ((r (replace-lenition word)))
	       (when r
		 (list
		  r
		  r
		  1
		  'D
		  'LM))))
	 
	 ;; Rzeczowniki 3. deklinacji
	 ;;  w dopeniaczu liczby pojedynczej wzoru twardego [12]
	 #'(lambda (word root ending-length)
	     (when (hard-ending root)
	       (list
		(string-concat root "a")
		(string-concat root "a")
		3
		'D
		'LP)))
	 ;;  w mianowniku liczby pojedynczej wzoru mikkiego
	 #'(lambda (word root ending-length)
	     (when (soft-ending root)
	       (list
		word
		word
		3
		'M
		'LP)))
	 ;;  w dopeniaczu, celowniku i miejscowniku liczby pojedynczej
	 ;;  wzoru mikkiego [12]
	 ;;; a -- czytelnia, ziemia
	 #'(lambda (word root ending-length)
	     (when (soft-ending word)
	       (list
		(string-concat word "a")
		(string-concat word "a")
		3
		'(D C Mc)
		'LP)))
	 ;;; b -- linia
	 #'(lambda (word root ending-length)
	     (when (soft-ending root)
	       (list
		(string-concat root "a")
		(string-concat root "a")
		3
		'(D C Mc)
		'LP)))
	 ;;; c -- gospodyni, pani
	 #'(lambda (word root ending-length)
	     (when (soft-ending word)
	       (list
		word
		word
		3
		'(D C Mc)
		'LP)))
	 ;;  w mianowniku, bierniku, woaczu liczby mnogiej wzoru twardego
	 ;;  [15]
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "k")
		       (match-ending root "g"))
	       (list
		(string-concat root "a")
		(string-concat root "a")
		3
		'(M B W)
		'LM)))
	 ;;  w dopeniaczu liczby mnogiej wzoru mikkiego [18]
	 #'(lambda (word root ending-length)
	     (let ((r (if (match-ending root "i")
			  (string-concat root "a")
			(string-concat word "a"))))
	       (list
		r
		r
		3
		'D
		'LM)))
	      
	 ;; Rzeczowniki 4. deklinacji
	 ;;  w dopeniaczu, celowniku, woaczu i miejscowniku liczby
	 ;;  pojedynczej [19]
	 #'(lambda (word root ending-length)
	     (let ((r (cond
		       ((consonant-ending root)
			(replace-lenition word))
		       ((match-ending root "e")
			(string-append root "j")))))
	       (when r
		 (list
		  r
		  r
		  4
		  '(D C W Mc)
		  'LP))))
	 ;;  w mianowniku, bierniku, woaczu i dopeniaczu liczby mnogiej
	 ;;  [20] oraz [19]
	 ;;  UWAGA: [20] bardzo mnie niepokoi...
	 #'(lambda (word root ending-length)
	     (let ((r (or (replace-lenition word) root)))
	       (list
		r
		r
		4
		'(M B W D)
		'LM)))) 

   (list "y"
	 ;; Rzeczowniki 1. deklinacji:
	 ;;  w mianowniku, bierniku i woaczu liczby mnogiej wzoru twardego (forma
	 ;;  niemskoosobowa)
	 #'(lambda (word root ending-length)
	     (let ((x (or (find-1dcl-root root) root)))
	       (list
		root
		x  
		1
		'(M B W)
		'LM)))
	 ;;  w mianowniku i woaczu liczby mnogiej wzoru mikkiego
	 ;;  (tylko rzeczowniki osobowe z przyrostkiem -ec)
	 #'(lambda (word root ending-length)
	     (when (and (match-ending root "c") (not (find-1dcl-root root)))
	       (list
		(string-concat (remove-ending root 1) "iec")
		(string-concat (remove-ending root 1) "iec")
		1
		'(M W)
		'LM)))
	 ;;  w dopeniaczu liczby mnogiej wzoru mikkiego po spgoskach
	 ;;  stwardniaych
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      1
	      'D
	      'LM))
	 
	 ;; Rzeczowniki 3. deklinacji
	 ;;  w dopeniaczu liczby pojedynczej wzoru twardego [12]
	 ;;  w dopeniaczu, celowniku i miejscowniku wzoru twardego
	 ;;  UWAGA: Na razie nie da si rozrni te dwa wzory a wic:
	 #'(lambda (word root ending-length)
	     (list
	      (string-concat root "a")
	      (string-concat root "a")
	      3
	      '(D C Mc)
	      'LP))
	 ;;  w mianowniku, bierniku i woaczu liczby mnogiej wzoru twardego
	 ;;  [15]
	 #'(lambda (word root ending-length)
	     (list
	      (string-concat root "a")
	      (string-concat root "a")
	      3
	      '(M B W)
	      'LM))
	 
	 ;; Rzeczowniki 4. deklinacji
	 ;;  w dopeniaczu, celowniku, woaczu i miejscowniku liczby
	 ;;  pojedynczej [19]
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      4
	      '(D C W Mc)
	      'LP))
	 ;;  w mianowniku, bierniku, woaczu i dopeniaczu[19]
	 ;;  liczby mnogiej [20]
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      4
	      '(M B W D)
	      'LM)))

   (list "w"
	 ;; Rzeczowniki 1. deklinacji
	 ;;  w dopeniaczu i bierniku liczby mnogiej
	 #'(lambda (word root ending-length)
	     (let ((x (or (find-1dcl-root root) root)))
	       (list
		root
		x
		1
		'(D B)
		'LM)))
	 ;;  w dopeniaczu liczby mnogiej wzoru mikkiego rzeczownikw z
	 ;;  przyrostkiem -ec
	 #'(lambda (word root ending-length)
	     (when (and (match-ending root "c") (not (find-1dcl-root root)))
	       (list
		(string-concat root "iec")
		(string-concat root "iec")
		1
		'D
		'LM)))
	 ;; Rzeczowniki 2. deklinacji na -um w dopeniaczu liczby mnogiej [10]
	 #'(lambda (word root ending-length)
	     (list
	      root 
	      (string-concat root "um")
	      2
	      'D
	      'LM)))

   (list "om"
	 ;; Rzeczowniki wszystkich deklinacji w celowniku liczby mnogiej
	 ;;; 1. dekl.
	 #'(lambda (word root ending-length)
	     (let ((x (or (find-1dcl-root root) root)))
	       (list
		root
		(replace-lenition x)
		1
		'C
		'LM)))
	 
	 #'(lambda (word root ending-length)
	     (when (and (match-ending root "c") (not (find-1dcl-root root)))
	       (list
		root
		(string-concat (remove-ending root 1) "iec")
		1
		'C
		'LM)))
	 ;;; 2. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      2
	      'C
	      'LM))	     
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "on")
		       (match-ending root "t"))
	       (let ((r (string-concat
			 (remove-ending root 2)
			 "")))
		 (list
		  r
		  r
		  2
		  'C
		  'LM))))
	 ;;; 3. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      (string-concat root "a")
	      (string-concat root "a")
	      3
	      'C
	      'LM))
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      3
	      'C
	      'LM))
	 ;;; 4. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      (replace-lenition root)
	      (replace-lenition root)
	      4
	      'C
	      'LM)))
   
   (list "ach"
	 ;; Rzeczowniki wszystkich deklinacji w miejscowniku liczby mnogiej
	 ;;; 1. dekl.
	 #'(lambda (word root ending-length)
	     (let ((x (or (find-1dcl-root root) root)))
	       (list
		root
		(replace-lenition x)
		1
		'Mc
		'LM)))
	 
	 #'(lambda (word root ending-length)
	     (when (and (match-ending root "c") (not (find-1dcl-root root)))
	       (list
		root
		(string-concat (remove-ending root 1) "iec")
		1
		'Mc
		'LM)))
	 ;;; 2. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      2
	      'Mc
	      'LM))	     
	 #'(lambda (word root ending-length)
	     (when (or (match-ending root "on")
		       (match-ending root "t"))
	       (let ((r (string-concat
			 (remove-ending root 2)
			 "")))
		 (list
		  r
		  r
		  2
		  'Mc
		  'LM))))
	 ;;; 3. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      (string-concat root "a")
	      (string-concat root "a")
	      3
	      'Mc
	      'LM))
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      3
	      'Mc
	      'LM))
	 ;;; 4. dekl.
	 #'(lambda (word root ending-length)
	     (list
	      (replace-lenition root)
	      (replace-lenition root)
	      4
	      'Mc
	      'LM)))

   (list "o"
	 ;; Rzeczowniki 2. deklinacji w mianowniku, bierniku i woaczu liczby
	 ;; pojedynczej
	 #'(lambda (word root ending-length)
	     (list
	      root
	      word
	      2
	      '(M B W)
	      'LP))
	 ;; Rzeczowniki 3. deklinacji
	 ;;   w mianowniku liczby pojedynczej
	 ;;   wzoru twardego [nazwiska mczyzn]
	 ;;   UWAGA: Na razie s ignorowane.

	 ;;   w woaczu liczby pojedynczej
	 ;;; a - ziemia, ulica
	 #'(lambda (word root ending-length)
	     (list
	      root
	      (string-concat root "a")
	      3
	      'W
	      'LP))
	 ;;; b - gospodyni, pani
	 #'(lambda (word root ending-length)
	     (when (match-ending root "i")
	       (list
		root
		root
		3
		'W
		'LP))))

   (list ""
	 ;; Rzeczowniki 2. deklinacji.
	 ;;  w mianowniku, bierniku i woaczu liczby pojedynczej [9]
	 #'(lambda (word root ending-length)
	     (list
	      (string-concat root "en")
	      word
	      2
	      '(M B W)
	      'LP))
	  
	 ;; Rzeczowniki 3. deklinacji
	 ;;  w bierniku liczby pojedynczej (p. [13])
	 #'(lambda (word root ending-length)
	     (list
	      root
	      (string-concat root "a")
	      3
	      'B
	      'LP))
	 #'(lambda (word root ending-length)
	     (list
	      root
	      root
	      3
	      'B
	      'LP)))

   (list ""
	 ;; Rzeczowniki 3. deklinacji
	 ;;  w narzdniku liczby pojedynczej wzoru twardego
	 #'(lambda (word root ending-length)
	     (when (hard-ending root)
	       (list
		root
		(string-concat root "a")
		3
		'N
		'LP)))
	 ;;  w bierniku i narzdniku liczby pojedynczej wzoru mikkiego [13]
	 #'(lambda (word root ending-length)
	     (when (soft-ending root)
	       (list
		root
		(string-concat root "a")
		3
		'(B N)
		'LP)))
	 #'(lambda (word root ending-length)
	     (when (soft-ending root)
	       (list
		root
		root
		3
		'(B N)
		'LP)))
	 
	 ;; Rzeczowniki 4. deklinacji
	 ;;  w narzdniku liczby pojedynczej
	 #'(lambda (word root ending-length)
	     (let ((r (or (replace-lenition root) root)))
	       (list
		r
		r
		4
		'N
		'LP)))) ))

(defun analyze-substantive (word)
  "Zwraca list moliwych form rzeczownika WORD. Poprawne formy trzeba
wybiera wedug sownika"
  (when (> (length word) 2)
    (analyze-word word substantive-drivers)))
		
;; Sownik. Format:
;;   (list Rzeczownik Deklinacja)
;;   Szczegy zale od czi mowy. Dla rzeczownikw jest to numer deklinacji.
;;   W przyszoci lista moe mie 3-ci element -- semantyczn definicj

(defvar subst-dictionary nil)

(defmacro dict-word (entry)
  `(nth 0 ,entry))

(defmacro dict-decl (entry)
  `(nth 1 ,entry))

;; Sownik form-wyjtkw.
(defvar subst-exceptions nil)

(defun dict-lookup-substantive (flist)
  (mapcan
   #'(lambda (form)
       (let ((dict (subst-get-dict form))
	     (decl (subst-get-decl form)))
	 (dolist (x subst-dictionary)
	   (when (and (string-equal dict (dict-word x))
		      (= decl (dict-decl x)))
	     (return (list form))))))
   flist))

(defun lookup-substantive (flist orig)
  (or (mapcan
       #'(lambda (x)
	   (if (string-equal (car x) orig)
	       (list (cdr x))))
       subst-exceptions)
      (dict-lookup-substantive flist)))
		  
(defun learn-substantive (flist orig)
  (when (not (null flist))
    (while t
      (let ((ans
	     (get-list-answer
	      "Ktra z odpowiedzi jest prawidowa (numer, lista numerw, <CR> jeli adna)?")))

	(cond
	 ((listp ans)
	  (unless (dolist (x ans)
		    (cond
		     ((not (numberp x))
		      (format t "Nie liczba: ~A. Sprbuj jeszcze raz.~%" x)
		      (return t))
		     ((>= x (length flist))
		      (format t "Nie ma odpowiedzi ~A~%" x)
		      (return t))))
	    (mapcar
	     #'(lambda (n)
		 (let ((elem (nth n flist)))
		   (dict-store subst-dictionary (list (subst-get-dict elem)
						      (subst-get-decl elem)))))
	     ans)
	    (return)))
	 ((or (not ans) (and (stringp ans) (string= ans "")))
	  (return))
	 (t
	  (format t "Niepoprawna odpowied.~%"))))))

  (while (getyn "Nowa definicja (Tak/Nie)?")
    (format t "| Definicja formy `~A'~%" orig)
    (let (decl root word number case)
      (setq word (getstring "| Mianownik:"))
      (unless (or (null word) (string= word ""))
	(setq root (getstring "| Rdze:"))
	(setq decl (while t
		     (let ((n (getnumber "| Deklinacja:")))
		       (if (or (> n 4) (< n 1))
			   (format t "| BD: Niepoprawna deklinacja.~%")
			 (return n)))))
	(setq number (while t
		       (let ((s (getsymbol "| Liczba:")))
			 (if (or (eq s 'LP) (eq s 'LM))
			     (return s)
			   (format t
				   "| BD: Niepoprawna liczba. Podaj LP albo LM.~%")))))
	(setq case (while t
		     (let ((cs (get-list-answer "| Przypadki:")))
		       (unless (dolist (x cs)
				 (unless (assoc x case-table)
				   (format t "| Niepoprawny przypadek: ~A~%" x)
				   (return t)))
			 (return cs))))))

      (let ((dfn (list root word decl number case)))
	(format t "|| Definicja: ")
	(print-result dfn)
	(when (getyn "|| Doda do sownika?")
	  (unless (dict-lookup-substantive (list dfn))
	    (dict-store subst-dictionary (list word decl)))
	  (dict-store subst-exceptions (cons orig dfn)))))))


;; Sprawdzanie (i ewentualnie samonauczanie)

(defun guess-declination (word)
  (list
   word
   word
   (cond
    ((hard-ending word) 1)
    ((or (match-ending word "o")
	 (match-ending word "e")
	 (match-ending word "um"))
     2)
    ((or (match-ending word "a")
	 (match-ending word "i"))
     3)
    ((soft-ending word) 4)
    (t
     4)) ;; FIXME
   'M
   'LP))

(defun get-substantive (word)
  (let ((r (or (analyze-substantive word)
	       (list (guess-declination word)))))
    (lookup-substantive r word)))

(defun analyze-and-learn-substantive (word)
  (let ((r (or (analyze-substantive word)
	       (list (guess-declination word)))))
    (format t "Analiza 'a priori':~%")
    (let ((n 0))
      (mapcar #'(lambda (x)
		  (format t "~A: " n)
		  (setq n (1+ n))
		  (print-result x))
	      r))
	
    (let ((res (lookup-substantive r word)))
      (cond
       (res
	(format t "Analiza a posteriori:~%")
	(mapcar #'(lambda (x)
		    (print-result x))
		res))
       (t
	(learn-substantive r word))))))

(defun read-print-substantive ()
  (format t "~&Rzeczownik: ")
  (let ((line (read-line)))
    (when (and line (string-not-equal line ""))
      (analyze-and-learn-substantive line)
      t)))

(defun check-substantives ()
  (while (read-print-substantive)))

;; Sprawdzanie
;(load "util.lisp")
;(load-dictionaries)
;(check-substantives)
;(save-dictionaries)
;(quit)
		
